/* Low-level BerkeleyDB functions */

/* DB4.1 and DB4.0 cross-compatibility issues */
int
sp_db_open(DB *db, DB_TXN *txnid, const char *file, const char *database, DBTYPE type, u_int32_t flags, int mode)
{
#if (DB_VERSION_MAJOR == 4 && DB_VERSION_MINOR >= 1)
    return db->open(db, txnid, file, database, type, flags, mode);
#else
    return db->open(db, file, database, type, flags, mode);
#endif /* (DB_VERSION_MAJOR == 4 && DB_VERSION_MINOR >= 1) */
}

/* db_lookup_entry() - lookup a db4 hash entry, return the value */
int
db_lookup_entry(char *dbfile, char *key, char *value)
{
    DB *db;
    DBT dbk, dbv;
    int dbfd;
    struct flock lockinfo;
    int rc;
    
    if(dbfile == NULL || key == NULL) return 0;

    memset(&dbk, '\0', sizeof(DBT));
    memset(&dbv, '\0', sizeof(DBT));
    memset(&lockinfo, '\0', sizeof(lockinfo));

    lockinfo.l_type = F_RDLCK;

    if((rc = db_create(&db, NULL, 0)) != 0) {
        spasm_error(LOG_ERR, "(db_lookup_entry) Error creating db: %s", db_strerror(rc));
        return 0;
    }
    if((rc = sp_db_open(db, NULL, dbfile, NULL, DB_HASH, DB_RDONLY, 0)) != 0) {
        if(rc != 2) spasm_error(LOG_ERR, "(db_lookup_entry) Error opening database %s: %s", dbfile, db_strerror(rc));
        db->close(db, 0);
        return 0;
    }
    if((rc = db->fd(db, &dbfd)) != 0) {
        spasm_error(LOG_ERR, "(db_lookup_entry) Error reading file descriptor for database: %s", db_strerror(rc));
        db->close(db, 0);
        return 0;
    }
    if(fcntl(dbfd, F_SETLKW, &lockinfo) == -1) {
        spasm_error(LOG_ERR, "(db_lookup_entry) Error locking database: %m");
        db->close(db, 0);                                                                                                                             
        return 0;                                                                                                                                     
    }
    dbk.data = key;
    dbk.size = strlen(key);
    if((rc = db->get(db, NULL, &dbk, &dbv, 0)) != 0) {
        if(rc != DB_NOTFOUND) spasm_error(LOG_ERR, "(db_lookup_entry) Error getting value from database: %s", db_strerror(rc));
        db->close(db, 0);
        return 0;
    }
    if(value != NULL && dbv.size && dbv.size <= SPASM_DBT_MAX)
        strncpy(value, (char *)dbv.data, dbv.size);
    db->close(db, 0);
    return 1;
}

/* db_insert_entry() - insert a db4 hash entry */
void
db_insert_entry(char *dbfile, char *key, char *value)
{     
    DB *db;
    DBT dbk, dbv;
    int dbfd;
    struct flock lockinfo;
    int rc;
    
    if(dbfile == NULL || key == NULL || value == NULL) return;
              
    memset(&dbk, '\0', sizeof(DBT));
    memset(&dbv, '\0', sizeof(DBT));
    memset(&lockinfo, '\0', sizeof(lockinfo));
    
    lockinfo.l_type = F_WRLCK;
    
    if((rc = db_create(&db, NULL, 0)) != 0) {
        spasm_error(LOG_ERR, "(db_insert_entry) Error creating db: %s", db_strerror(rc));
        return;
    }
    if((rc = sp_db_open(db, NULL, dbfile, NULL, DB_HASH, DB_CREATE, 0644)) != 0) {
        spasm_error(LOG_ERR, "(db_insert_entry) Error opening database: %s", db_strerror(rc));
        db->close(db, 0);
        return;
    }
    if((rc = db->fd(db, &dbfd)) != 0) {
        spasm_error(LOG_ERR, "(db_insert_entry) Error reading file descriptor for database: %s", db_strerror(rc));
        db->close(db, 0);
        return;
    }
    if(fcntl(dbfd, F_SETLKW, &lockinfo) == -1) {
        spasm_error(LOG_ERR, "(db_insert_entry) Error locking database: %m");
        db->close(db, 0);
        return;
    }   
    dbk.data = key;
    dbk.size = strlen(key);
    dbv.data = value;
    dbv.size = strlen(value);
    if((rc = db->put(db, NULL, &dbk, &dbv, 0)) != 0) {
        spasm_error(LOG_ERR, "(db_insert_entry) Error adding %s/%s to %s: %s", key, value, dbfile, db_strerror(rc));
    }
    if((rc = db->close(db, 0))) {
        spasm_error(LOG_ERR, "(db_insert_entry) Error closing database: %s", db_strerror(rc));
        if(close(dbfd) != 0) spasm_error(LOG_ERR, "(db_insert_entry) Error closing dbfd: %m");
    }
    return;
}

/* db_delete_entry() - delete a db4 entry */
void
db_delete_entry(char *dbfile, char *key)
{     
    DB *db;
    DBT dbk;
    int dbfd;
    int rc;
    struct flock lockinfo;
    
    if(dbfile == NULL || key == NULL) return;
    memset(&dbk, '\0', sizeof(DBT));
    memset(&lockinfo, '\0', sizeof(lockinfo));
    lockinfo.l_type = F_WRLCK;
    if((rc = db_create(&db, NULL, 0)) != 0) {
        spasm_error(LOG_ERR, "(db_delete_entry) Error creating db: %s", db_strerror(rc));
        return;
    }
    if((rc = sp_db_open(db, NULL, dbfile, NULL, DB_HASH, DB_CREATE, 0644)) != 0) {
        spasm_error(LOG_ERR, "(db_delete_entry) Error opening database: %s", db_strerror(rc));
        db->close(db, 0);
        return;
    }
    if((rc = db->fd(db, &dbfd)) != 0) {
        spasm_error(LOG_ERR, "(db_delete_entry) Error reading file descriptor for database: %s", db_strerror(rc));
        db->close(db, 0);
        return;
    }
    if(fcntl(dbfd, F_SETLKW, &lockinfo) == -1) {
        spasm_error(LOG_ERR, "(db_delete_entry) Error locking database: %m");
        db->close(db, 0);
        return;
    }   
    dbk.data = key;
    dbk.size = strlen(key);
    if((rc = db->del(db, NULL, &dbk, 0)) != 0) {
        spasm_error(LOG_ERR, "(db_delete_entry) Error deleting %s from %s: %s", key, dbfile, db_strerror(rc));
    }
    if((rc = db->close(db, 0)) != 0) {
        spasm_error(LOG_ERR, "(db_delete_entry) Error closing database: %s", db_strerror(rc));
        if(close(dbfd) != 0) spasm_error(LOG_ERR, "(db_delete_entry) Error closing dbfd: %m");
    }
    return;
}

/* db_dump_entries() - dump spasm_list of entries in db4 database */
struct spasm_list*
db_dump_entries(char *dbfile, int vals)
{   
    struct spasm_list *entries = NULL;
    struct spasm_list *first = NULL;
    DB *db;
    DBT dbk, dbv;
    DBC *dbcp;                                                                                                                                         
    int dbfd, rc;                                                                                                                                      

    if(dbfile == NULL) return NULL;
    memset(&dbk, '\0', sizeof(DBT));
    memset(&dbv, '\0', sizeof(DBT));

    if((rc = db_create(&db, NULL, 0)) != 0) {
        spasm_error(LOG_ERR, "(db_dump_entries) Error creating db: %s", db_strerror(rc));
        return NULL;
    }
    if((rc = sp_db_open(db, NULL, dbfile, NULL, DB_HASH, DB_RDONLY, 0)) != 0) {
        if(rc != 2) spasm_error(LOG_ERR, "(db_dump_entries) Error opening database (%d): %s", rc, db_strerror(rc));
        db->close(db, 0);
        return NULL;
    }
    if((rc = db->fd(db, &dbfd)) != 0) {
        spasm_error(LOG_ERR, "(db_dump_entries) Error reading file descriptor for database: %s", db_strerror(rc));
        db->close(db, 0);
        return NULL;
    }
    if((rc = db->cursor(db, NULL, &dbcp, 0)) != 0) {
        spasm_error(LOG_ERR, "(db_dump_entries) Error retrieving db cursor: %s", db_strerror(rc));
        db->close(db, 0);
        return NULL;
    }
    while(dbcp->c_get(dbcp, &dbk, &dbv, DB_NEXT) == 0) {
        if(first == NULL) {
            if((first = calloc(sizeof(struct spasm_list), 1)) != NULL) {
                if(dbk.size <= SPASM_DBT_MAX) {
                    entries = first;
                    entries->next = NULL;
                    strncpy(entries->entry, dbk.data, dbk.size);
                    if(vals) {
                        if(dbv.size <= SPASM_DBT_MAX) strncpy(entries->strval, dbv.data, dbv.size);
                        else spasm_error(LOG_ERR, "(db_dump_entries) Value is too large");
                    }
                } else spasm_error(LOG_ERR, "(db_dump_entries) Entry is too large");
            } else spasm_error(LOG_ERR, "(db_dump_entries) Error allocating memory for first: %m");
        } else {
            if((entries->next = calloc(sizeof(struct spasm_list) + 1, 1)) != NULL) {
                entries = entries->next;
                entries->next = NULL;
                if(dbk.size <= SPASM_DBT_MAX) {
                    strncpy(entries->entry, dbk.data, dbk.size);
                    if(vals) {
                        if(dbv.size <= SPASM_DBT_MAX) strncpy(entries->strval, dbv.data, dbv.size);                                                  
                        else spasm_error(LOG_ERR, "(db_dump_entries) Value is too large");
                    }
                } else spasm_error(LOG_ERR, "(db_dump_entries) Entry is too large");
            } else spasm_error(LOG_ERR, "(db_dump_entries) Error allocating memory for entries->next: %m");
        }
    }
    dbcp->c_close(dbcp);
    db->close(db, 0);
    return first;
}
